package org.apache.bcel.util;

package .var.folders._j.vfbv187d0bb8b8w5spzgg8180000gn.T;

import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.nio.charset.StandardCharsets;
import java.util.Locale;
import org.apache.bcel.Const;
import org.apache.bcel.Repository;
import org.apache.bcel.classfile.ClassParser;
import org.apache.bcel.classfile.ConstantValue;
import org.apache.bcel.classfile.EmptyVisitor;
import org.apache.bcel.classfile.Field;
import org.apache.bcel.classfile.JavaClass;
import org.apache.bcel.classfile.Method;
import org.apache.bcel.classfile.Utility;
import org.apache.bcel.generic.ArrayType;
import org.apache.bcel.generic.ConstantPoolGen;
import org.apache.bcel.generic.MethodGen;
import org.apache.bcel.generic.Type;
import org.apache.bcel.util.BCELFactory;
import org.apache.commons.lang3.StringUtils;

public class BCELifier extends EmptyVisitor {
  private static final String BASE_PACKAGE = Const.class.getPackage().getName();
  
  private static final String CONSTANT_PREFIX = Const.class.getSimpleName() + ".";
  
  private final JavaClass clazz;
  
  private final PrintWriter printWriter;
  
  private final ConstantPoolGen constantPoolGen;
  
  static JavaClass getJavaClass(String name) throws ClassNotFoundException, IOException {
    JavaClass javaClass;
    if ((javaClass = Repository.lookupClass(name)) == null)
      javaClass = (new ClassParser(name)).parse(); 
    return javaClass;
  }
  
  public static void main(String[] argv) throws Exception {
    if (argv.length != 1) {
      System.out.println("Usage: BCELifier className");
      System.out.println("\tThe class must exist on the classpath");
      return;
    } 
    org.apache.bcel.util.BCELifier bcelifier = new org.apache.bcel.util.BCELifier(getJavaClass(argv[0]), System.out);
    bcelifier.start();
  }
  
  static String printArgumentTypes(Type[] argTypes) {
    if (argTypes.length == 0)
      return "Type.NO_ARGS"; 
    StringBuilder args = new StringBuilder();
    for (int i = 0; i < argTypes.length; i++) {
      args.append(printType(argTypes[i]));
      if (i < argTypes.length - 1)
        args.append(", "); 
    } 
    return "new Type[] { " + args.toString() + " }";
  }
  
  static String printFlags(int flags) {
    return printFlags(flags, FLAGS.UNKNOWN);
  }
  
  public static String printFlags(int flags, FLAGS location) {
    if (flags == 0)
      return "0"; 
    StringBuilder buf = new StringBuilder();
    for (int i = 0, pow = 1; pow <= 32768; i++) {
      if ((flags & pow) != 0)
        if (pow == 32 && location == FLAGS.CLASS) {
          buf.append(CONSTANT_PREFIX).append("ACC_SUPER | ");
        } else if (pow == 64 && location == FLAGS.METHOD) {
          buf.append(CONSTANT_PREFIX).append("ACC_BRIDGE | ");
        } else if (pow == 128 && location == FLAGS.METHOD) {
          buf.append(CONSTANT_PREFIX).append("ACC_VARARGS | ");
        } else if (i < Const.ACCESS_NAMES_LENGTH) {
          buf.append(CONSTANT_PREFIX).append("ACC_").append(Const.getAccessName(i).toUpperCase(Locale.ENGLISH)).append(" | ");
        } else {
          buf.append(String.format(CONSTANT_PREFIX + "ACC_BIT %x | ", new Object[] { Integer.valueOf(pow) }));
        }  
      pow <<= 1;
    } 
    String str = buf.toString();
    return str.substring(0, str.length() - 3);
  }
  
  static String printType(String signature) {
    Type type = Type.getType(signature);
    byte t = type.getType();
    if (t <= 12)
      return "Type." + Const.getTypeName(t).toUpperCase(Locale.ENGLISH); 
    if (type.toString().equals("java.lang.String"))
      return "Type.STRING"; 
    if (type.toString().equals("java.lang.Object"))
      return "Type.OBJECT"; 
    if (type.toString().equals("java.lang.StringBuffer"))
      return "Type.STRINGBUFFER"; 
    if (type instanceof ArrayType) {
      ArrayType at = (ArrayType)type;
      return "new ArrayType(" + printType(at.getBasicType()) + ", " + at.getDimensions() + ")";
    } 
    return "new ObjectType(\"" + Utility.signatureToString(signature, false) + "\")";
  }
  
  static String printType(Type type) {
    return printType(type.getSignature());
  }
  
  public BCELifier(JavaClass clazz, OutputStream out) {
    this.clazz = clazz;
    this.printWriter = new PrintWriter(new OutputStreamWriter(out, StandardCharsets.UTF_8), false);
    this.constantPoolGen = new ConstantPoolGen(this.clazz.getConstantPool());
  }
  
  private void printCreate() {
    this.printWriter.println("  public void create(OutputStream out) throws IOException {");
    Field[] fields = this.clazz.getFields();
    if (fields.length > 0)
      this.printWriter.println("    createFields();"); 
    Method[] methods = this.clazz.getMethods();
    for (int i = 0; i < methods.length; i++)
      this.printWriter.println("    createMethod_" + i + "();"); 
    this.printWriter.println("    _cg.getJavaClass().dump(out);");
    this.printWriter.println("  }");
    this.printWriter.println();
  }
  
  private void printMain() {
    String className = this.clazz.getClassName();
    this.printWriter.println("  public static void main(String[] args) throws Exception {");
    this.printWriter.println("    " + className + "Creator creator = new " + className + "Creator();");
    this.printWriter.println("    creator.create(new FileOutputStream(\"" + className + ".class\"));");
    this.printWriter.println("  }");
  }
  
  public void start() {
    visitJavaClass(this.clazz);
    this.printWriter.flush();
  }
  
  public void visitField(Field field) {
    this.printWriter.println();
    this.printWriter.println("    field = new FieldGen(" + 
        printFlags(field.getAccessFlags()) + ", " + printType(field.getSignature()) + ", \"" + field.getName() + "\", _cp);");
    ConstantValue cv = field.getConstantValue();
    if (cv != null)
      this.printWriter.println("    field.setInitValue(" + cv + ")"); 
    this.printWriter.println("    _cg.addField(field.getField());");
  }
  
  public void visitJavaClass(JavaClass clazz) {
    String className = clazz.getClassName();
    String superName = clazz.getSuperclassName();
    String packageName = clazz.getPackageName();
    String inter = Utility.printArray((Object[])clazz.getInterfaceNames(), false, true);
    if (StringUtils.isNotEmpty(inter)) {
      className = className.substring(packageName.length() + 1);
      this.printWriter.println("package " + packageName + ";");
      this.printWriter.println();
    } 
    this.printWriter.println("import " + BASE_PACKAGE + ".generic.*;");
    this.printWriter.println("import " + BASE_PACKAGE + ".classfile.*;");
    this.printWriter.println("import " + BASE_PACKAGE + ".*;");
    this.printWriter.println("import java.io.*;");
    this.printWriter.println();
    this.printWriter.println("public class " + className + "Creator {");
    this.printWriter.println("  private InstructionFactory _factory;");
    this.printWriter.println("  private ConstantPoolGen    _cp;");
    this.printWriter.println("  private ClassGen           _cg;");
    this.printWriter.println();
    this.printWriter.println("  public " + className + "Creator() {");
    this.printWriter.println("    _cg = new ClassGen(\"" + (packageName.isEmpty() ? className : (packageName + "." + className)) + "\", \"" + superName + "\", \"" + clazz
        .getSourceFileName() + "\", " + printFlags(clazz.getAccessFlags(), FLAGS.CLASS) + ", new String[] { " + inter + " });");
    this.printWriter.println("    _cg.setMajor(" + clazz.getMajor() + ");");
    this.printWriter.println("    _cg.setMinor(" + clazz.getMinor() + ");");
    this.printWriter.println();
    this.printWriter.println("    _cp = _cg.getConstantPool();");
    this.printWriter.println("    _factory = new InstructionFactory(_cg, _cp);");
    this.printWriter.println("  }");
    this.printWriter.println();
    printCreate();
    Field[] fields = clazz.getFields();
    if (fields.length > 0) {
      this.printWriter.println("  private void createFields() {");
      this.printWriter.println("    FieldGen field;");
      for (Field field : fields)
        field.accept(this); 
      this.printWriter.println("  }");
      this.printWriter.println();
    } 
    Method[] methods = clazz.getMethods();
    for (int i = 0; i < methods.length; i++) {
      this.printWriter.println("  private void createMethod_" + i + "() {");
      methods[i].accept(this);
      this.printWriter.println("  }");
      this.printWriter.println();
    } 
    printMain();
    this.printWriter.println("}");
  }
  
  public void visitMethod(Method method) {
    MethodGen mg = new MethodGen(method, this.clazz.getClassName(), this.constantPoolGen);
    this.printWriter.println("    InstructionList il = new InstructionList();");
    this.printWriter.println("    MethodGen method = new MethodGen(" + printFlags(method.getAccessFlags(), FLAGS.METHOD) + ", " + printType(mg.getReturnType()) + ", " + 
        printArgumentTypes(mg.getArgumentTypes()) + ", new String[] { " + Utility.printArray((Object[])mg.getArgumentNames(), false, true) + " }, \"" + method
        .getName() + "\", \"" + this.clazz.getClassName() + "\", il, _cp);");
    this.printWriter.println();
    BCELFactory factory = new BCELFactory(mg, this.printWriter);
    factory.start();
    this.printWriter.println("    method.setMaxStack();");
    this.printWriter.println("    method.setMaxLocals();");
    this.printWriter.println("    _cg.addMethod(method.getMethod());");
    this.printWriter.println("    il.dispose();");
  }
}
